from typing import List, Optional, Union

import torch

from open_gpt.models.modeling import BaseModel


class LlamaModel(BaseModel):
    no_split_module_classes = ["LlamaDecoderLayer"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate(self, prompts: Union[str, List[str]], **kwargs):
        """Generate text from the given prompt."""

        prompts = [prompts] if isinstance(prompts, str) else prompts

        with torch.inference_mode():
            texts_outs = []

            for prompt in prompts:
                inputs = self.tokenizer(
                    prompt,
                    return_tensors="pt",
                )
                # Move inputs to the correct device
                for k, v in inputs.items():
                    if isinstance(v, torch.Tensor):
                        inputs[k] = v.to(self._device)

                inputs.pop('token_type_ids')

                outputs = self.model.generate(**inputs, **kwargs)[0].tolist()

                prompt_len = len(prompt)

                text = self.tokenizer.decode(
                    outputs,
                    clean_up_tokenization_spaces=True,
                    skip_special_tokens=True,
                )

                text = text[prompt_len:] if text[:prompt_len] == prompt else text
                texts_outs.append(text)

            return texts_outs



from pathlib import Path
from typing import List, Optional, Union

import torch


def create_model(
    model_name: str,
    precision: Optional[str] = None,
    device: Optional[Union[str, torch.device]] = None,
    device_map: Optional[Union[str, List[int]]] = None,
    **kwargs,
):
    """Create a model.

    :param model_name: The name of the model to create.
    :param precision: The precision to use. Can be one of ``"float16"``, ``"float32"``, ``"float64"``, ``"bfloat16"``, ``"mixed"`` or ``None``.
    :param device: The device to use. Can be one of ``"cpu"``, ``"cuda"``, ``"cuda:X"`` or ``None``.
    :param device_map: The device map to use. Can be one of ``"balanced"``, ``"single"`` or a list of device IDs.
    :param kwargs: Additional keyword arguments to pass to the model.
    """

    if model_name.startswith('facebook/llama') or model_name.startswith(
        'decapoda-research/llama'
    ):
        from .models.llama.modeling import LlamaModel

        return LlamaModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    elif model_name.startswith('lmsys/vicuna'):
        assert not model_name.endswith(
            'v0'
        ), 'You are using an outdated model, please use the newer version ``v1.1+``'
        from .models.vicuna.modeling import VicunaModel

        return VicunaModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    elif model_name.startswith('EleutherAI/pythia'):
        from .models.pythia.modeling import PythiaModel

        return PythiaModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    elif model_name.startswith('stabilityai/stablelm') or model_name.startswith(
        'CarperAI/stable-vicuna'
    ):
        from .models.stablelm.modeling import StableLMModel

        return StableLMModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    elif model_name.startswith('fnlp/moss'):
        from .models.moss.modeling import MossModel

        return MossModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    elif model_name.startswith('openflamingo/OpenFlamingo'):
        from .models.flamingo.modeling import FlamingoModel

        return FlamingoModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )
    else:
        from .models.modeling import BaseModel

        return BaseModel(
            model_name,
            device=device,
            precision=precision,
            device_map=device_map,
            **kwargs,
        )


def create_flow(
    model_name_or_path: str, protocol='http', port=51000, replicas: int = 1
):
    from jina import Flow

    if 'flamingo' in model_name_or_path:
        from .serve.executors.flamingo import FlamingoExecutor as Executor
    else:
        from serve.executors import CausualLMExecutor as Executor

    # normalize the model name to be used as flow executor name
    norm_name = model_name_or_path.split('/')[-1]
    norm_name = norm_name.replace('-', '_').lower()

    return Flow(protocol=protocol, port=port, cors=True).add(
        uses=Executor,
        uses_with={'model_name_or_path': model_name_or_path},
        name=f'{norm_name}_executor',
        replicas=replicas,
        timeout_ready=-1,
    )



from typing import List, Optional, Union

import torch
from PIL import Image

from open_gpt.models.modeling import BaseModel


class FlamingoModel(BaseModel):
    no_split_module_classes = ["LlamaDecoderLayer"]
    image_processor = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_model_and_transforms(
        self, model_name_or_path: str, tokenizer_name_or_path: Optional[str] = None
    ):
        from .loading import load_model_and_transforms

        self.model, self.tokenizer, self.image_processor = load_model_and_transforms(
            model_name_or_path,
            vision_model_name_or_path='ViT-L-14::openai',
            lang_model_name_or_path='facebook/llama-7b',
            tokenizer_name_or_path=tokenizer_name_or_path,
            dtype=self._dtype,
            device=self._device,
            device_map=self._device_map,
            no_split_module_classes=self.no_split_module_classes,
        )

        self.model.eval()

    def generate(self, prompt: str, inplace_images: List = [], **kwargs):
        """Generate text from the given prompt."""

        assert isinstance(prompt, str), "Prompt must be a string."
        prompt_len = len(prompt)

        with torch.inference_mode():
            vision_x = []

            for image in inplace_images:
                vision_x.append(
                    self.image_processor(
                        Image.open(image) if isinstance(image, str) else image
                    ).unsqueeze(0)
                )
            vision_x = torch.cat(vision_x, dim=0)
            vision_x = vision_x.unsqueeze(1).unsqueeze(0)

            lang_x = self.tokenizer(
                [prompt],
                padding=True,
                return_tensors="pt",
            )

            generated_tokens = self.model.generate(
                vision_inputs=vision_x,
                text_inputs=lang_x["input_ids"],
                attention_mask=lang_x["attention_mask"],
                max_new_tokens=20,
                num_beams=6,
                top_p=0.9,
                no_repeat_ngram_size=2,
                temperature=0.7,
                length_penalty=1.5,
            )[0].tolist()

            text = self.tokenizer.decode(
                generated_tokens,
                clean_up_tokenization_spaces=True,
                skip_special_tokens=True,
            )

            text = text[prompt_len:] if text[:prompt_len] == prompt else text

            return text



"""The serve module provides a simple way to serve a model using Jina."""

from jina import DocumentArray, Gateway


class SSEGateway(Gateway):
    """A simple SSE gateway that can be used to stream events to the client."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def setup_server(self):
        """Setup the server."""
        self.app = self._get_app()

        @self.app.get("/generate")
        async def generate(request):
            """Generate text."""

            async def generate_text():
                """Generate text."""
                async for message in self._generate(request):
                    yield message

            return await self._stream(generate_text())

    async def _generate(self, request):
        """Generate text."""
        docs = DocumentArray([{"text": request.args["text"][0]}])
        await self._post("/generate", docs)
        return docs



""" StableLM model.

Some code is copied from https://github.com/Stability-AI/StableLM

Original Apache 2.0 License

"""

from typing import List, Optional, Union

import torch
from transformers import StoppingCriteria, StoppingCriteriaList

from open_gpt.models.modeling import BaseModel


class StopOnTokens(StoppingCriteria):
    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs
    ) -> bool:
        stop_ids = [50278, 50279, 50277, 1, 0]
        for stop_id in stop_ids:
            if input_ids[0][-1] == stop_id:
                return True
        return False


class StableLMModel(BaseModel):
    """Wrapper for StableLM models by Stability AI.

    See https://github.com/Stability-AI/StableLM for more details.

    The quick way to use StableLM via :meth:`open_gpt.create_model`:

    ```python
    import open_gpt

    model = open_gpt.create_model('stabilityai/stablelm-tuned-alpha-7b')

    system_prompt = (
        '<|SYSTEM|># StableLM Tuned (Alpha version)\n'
        '- StableLM is a helpful and harmless open-source AI language model developed by StabilityAI.\n'
        '- StableLM is excited to be able to help the user, but will refuse to do anything that could be considered harmful to the user.\n'
        '- StableLM is more than just an information source, StableLM is also able to write poetry, short stories, and make jokes.\n'
        '- StableLM will refuse to participate in anything that could harm a human.'
    )


    prompt = f"{system_prompt}<|USER|>What's your mood today?<|ASSISTANT|>"


    # Generate text
    text_out = model.generate_text(prompt, max_length=50)
    ```

    :note: StableLM Tuned should be used with prompts formatted to <|SYSTEM|>...<|USER|>...<|ASSISTANT|>...


    ```python
    prompt = (
        '### Human: Write a Python script for text classification using Transformers and PyTorch\n'
        '### Assistant:\n'
    )

    # Generate text with StableLM-StableVicuna-13B
    model = open_gpt.create_model('CarperAI/stable-vicuna-13b-delta')
    ```
    """

    no_split_module_classes = ["GPTNeoXLayer"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def is_vicuna_model(self):
        return 'vicuna' in self._model_name_or_path

    def generate(self, prompts: Union[str, List[str]], **kwargs):
        """Generate text from the given prompt."""

        # patch to fix the issue with StableLM not stopping on
        return super().generate(
            prompts,
            stopping_criteria=StoppingCriteriaList([StopOnTokens()])
            if not self.is_vicuna_model
            else None,
            skip_special_tokens=False,
            **kwargs
        )



import os
import sys

from loguru import logger


def setup_logger():
    logging_level = os.environ.get("OPENGPT_LOG_LEVEL", "INFO")

    logger.remove()
    logger.add(
        sys.stdout,
        colorize=True,
        format=(
            "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
            "<level>{level: <8}</level> | <level>{message}</level>"
        ),
        level=logging_level,
    )
    logger.level("WARNING", color="<fg #d3d3d3>")
    return logger


logger = setup_logger()



from typing import TYPE_CHECKING, List, Optional, Union

import torch
from torch import nn

from ..helper import auto_dtype_and_device
from ..logging import logger
from .generation import GenerationMixin

if TYPE_CHECKING:
    from transformers import AutoModelForCausalLM, AutoTokenizer


class BaseModel(nn.Module, GenerationMixin):
    model: 'AutoModelForCausalLM'
    tokenizer: 'AutoTokenizer'
    no_split_module_classes: List[str] = None

    def __init__(
        self,
        model_name_or_path: str,
        tokenizer_name_or_path: Optional[str] = None,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[torch.device] = None,
        device_map: Optional[Union[str, List[int]]] = None,
        eval_mode: bool = True,
        **kwargs,
    ):
        """Create a model of the given name."""

        super().__init__()

        self._model_name_or_path = model_name_or_path

        self._dtype, self._device = auto_dtype_and_device(dtype, device)

        self._device_map = device_map
        if not self._device_map:
            logger.warning(
                f'To turn on tensor parallelism, set `device_map` to a list of GPU ids rather than `None`'
            )

        self._eval_mode = eval_mode

        self.load_model_and_transforms(
            model_name_or_path, tokenizer_name_or_path=tokenizer_name_or_path
        )

        self.post_init(**kwargs)

    def load_model_and_transforms(
        self, model_name_or_path: str, tokenizer_name_or_path: Optional[str] = None
    ):
        from .loading import load_model_and_tokenizer

        self.model, self.tokenizer = load_model_and_tokenizer(
            model_name_or_path,
            tokenizer_name_or_path=tokenizer_name_or_path,
            dtype=self._dtype,
            device=self._device,
            device_map=self._device_map,
            no_split_module_classes=self.no_split_module_classes,
        )

        # turn the eval mode off `eval_mode=False` in training
        if self._eval_mode:
            self.model.eval()

    def post_init(self, **kwargs):
        pass



from cleo.commands.command import Command
from cleo.helpers import argument, option


class ServeCommand(Command):
    name = "serve"

    description = "Start a model serving locally."

    arguments = [argument("model_name", "The name of the model to serve.")]
    options = [
        option(
            "protocol",
            None,
            "The protocol to serve the model on.",
            flag=False,
            default="http",
        ),
        option(
            "port", "p", "The port to serve the model on.", flag=False, default=51000
        ),
        option(
            "replicas", "r", "The number of replicas to serve.", flag=False, default=1
        ),
    ]

    help = """\
    This command allows you to start a model serving locally.
    
    To start a model serving locally, you can run:
        
        <comment>opengpt serve facebook/llama-7b</comment>"""

    def handle(self) -> int:
        from open_gpt.factory import create_flow

        with create_flow(
            self.argument('model_name'),
            protocol=self.option('protocol'),
            port=self.option('port'),
            replicas=self.option('replicas'),
        ) as flow:
            flow.block()

        return 0



"""This file contains a few functions to profile the memory usage of the model.

It is not meant to be used in production, but rather to help us debug the memory usage of the model.

The codes are borrowed from https://github.com/huggingface/accelerate/blob/main/benchmarks/measures_util.py
"""

import gc
import threading
import time

import psutil
import torch
from accelerate.utils import compute_module_sizes as _compute_module_sizes

KB = 1 << 10
MB = 1 << 20
GB = 1 << 30
T = 1e12


def compute_module_sizes(model):
    return _compute_module_sizes(model)


class PeakCPUMemory:
    def __init__(self):
        self.process = psutil.Process()
        self.peak_monitoring = False

    def peak_monitor(self):
        self.cpu_memory_peak = -1

        while True:
            self.cpu_memory_peak = max(
                self.process.memory_info().rss, self.cpu_memory_peak
            )

            # can't sleep or will not catch the peak right (this comment is here on purpose)
            if not self.peak_monitoring:
                break

    def start(self):
        self.peak_monitoring = True
        self.thread = threading.Thread(target=self.peak_monitor)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.peak_monitoring = False
        self.thread.join()
        return self.cpu_memory_peak


cpu_peak_tracker = PeakCPUMemory()


def start_measure():
    # Time
    measures = {"time": time.time()}

    gc.collect()
    torch.cuda.empty_cache()

    # CPU mem
    measures["cpu"] = psutil.Process().memory_info().rss
    cpu_peak_tracker.start()

    # GPU mem
    for i in range(torch.cuda.device_count()):
        measures[str(i)] = torch.cuda.memory_allocated(i)
    torch.cuda.reset_peak_memory_stats()

    return measures


def end_measure(start_measures):
    # Time
    measures = {"time": time.time() - start_measures["time"]}

    gc.collect()
    torch.cuda.empty_cache()

    # CPU mem
    measures["cpu"] = (
        psutil.Process().memory_info().rss - start_measures["cpu"]
    ) / 2**20
    measures["cpu-peak"] = (cpu_peak_tracker.stop() - start_measures["cpu"]) / 2**20

    # GPU mem
    for i in range(torch.cuda.device_count()):
        measures[str(i)] = (
            torch.cuda.memory_allocated(i) - start_measures[str(i)]
        ) / GB
        measures[f"{i}-peak"] = (
            torch.cuda.max_memory_allocated(i) - start_measures[str(i)]
        ) / GB

    return measures


def log_measures(measures, description):
    print(f"{description}:")
    print(f"- Time: {measures['time']:.2f}s")
    for i in range(torch.cuda.device_count()):
        print(f"- GPU {i} allocated: {measures[str(i)]:.3f}GiB")
        peak = measures[f"{i}-peak"]
        print(f"- GPU {i} peak: {peak:.3f}GiB")
    print(f"- CPU RAM allocated: {measures['cpu']:.3f}GiB")
    print(f"- CPU RAM peak: {measures['cpu-peak']:.3f}GiB")



"""The executor wraps the model and provides a simple way to run inference on the model."""

from multiprocessing.pool import ThreadPool
from typing import Dict, List, Optional, Union

import torch
from docarray import DocumentArray
from jina import Executor, requests

from open_gpt.factory import create_model
from open_gpt.logging import logger


class CausualLMExecutor(Executor):
    def __init__(
        self,
        model_name_or_path: str = '',
        minibatch_size: int = 1,
        device_map: Optional[Union[str, List[int]]] = None,
        precision: Optional[str] = None,
        num_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert (
            model_name_or_path
        ), '`model_name_or_path` must be provided to initialize the model and tokenizer.'

        self._model_name_or_path = model_name_or_path
        self._minibatch_size = minibatch_size
        self._thread_pool = ThreadPool(processes=num_workers)

        self.model = create_model(
            model_name_or_path, precision=precision, device_map=device_map, **kwargs
        )

        # warmup the model to avoid the first-time slowness
        self.model.generate(['Hello world!'])

    @requests(on='/generate')
    def generate(self, docs: 'DocumentArray', parameters: Dict = {}, **kwargs):
        prompted_da = DocumentArray(
            [d for d in docs if d.tags.get('prompt') or d.text is not None]
        )
        prompts = [d.tags['prompt'] or d.text for d in prompted_da]

        if prompts:
            result = self.model.generate(prompts, **parameters)
            for d, r in zip(prompted_da, result):
                d.tags['generated_text'] = r
        else:
            logger.warning('No prompts found in the request.')



import random
from typing import Optional, Union

import torch
import torch.nn as nn
from open_flamingo.src.helpers import GatedCrossAttentionBlock
from open_flamingo.src.utils import getattr_recursive, setattr_recursive

from ...helper import auto_dtype_and_device


class FlamingoLayer(nn.Module):
    def __init__(self, gated_cross_attn_layer, decoder_layer):
        super().__init__()
        self.gated_cross_attn_layer = gated_cross_attn_layer
        self.decoder_layer = decoder_layer
        self.vis_x = None
        self.media_locations = None
        self.device = self.decoder_layer.parameters().__next__().device

        # This is a hack to guarantee that the gated_cross_attn_layer is on the same device as the decoder_layer
        if self.gated_cross_attn_layer is not None:
            self.gated_cross_attn_layer.to(self.device)

    def is_conditioned(self) -> bool:
        """Check whether the layer is conditioned."""
        return self.vis_x is not None

    # Used this great idea from this implementation of Flamingo (https://github.com/dhansmair/flamingo-mini/)
    def condition_vis_x(self, vis_x):
        self.vis_x = vis_x.to(self.device) if vis_x is not None else vis_x

    def condition_media_locations(self, media_locations):
        self.media_locations = (
            media_locations.to(self.device)
            if media_locations is not None
            else media_locations
        )

    def condition_attend_previous(self, attend_previous):
        self.attend_previous = attend_previous

    def forward(
        self,
        lang_x,
        attention_mask=None,
        **decoder_layer_kwargs,
    ):
        if self.gated_cross_attn_layer is None:
            return self.decoder_layer(
                lang_x, attention_mask=attention_mask, **decoder_layer_kwargs
            )

        if self.vis_x is None:
            raise ValueError("vis_x must be conditioned before forward pass")

        if self.media_locations is None:
            raise ValueError("media_locations must be conditioned before forward pass")

        lang_x = lang_x.to(self.device)
        lang_x = self.gated_cross_attn_layer(
            lang_x,
            self.vis_x,
            media_locations=self.media_locations,
            attend_previous=self.attend_previous,
        )
        lang_x = self.decoder_layer(
            lang_x, attention_mask=attention_mask, **decoder_layer_kwargs
        )
        return lang_x


class FlamingoLMMixin(nn.Module):
    """
    Mixin to add cross-attention layers to a language model.
    """

    def set_decoder_layers_attr_name(self, decoder_layers_attr_name):
        self.decoder_layers_attr_name = decoder_layers_attr_name

    def _get_decoder_layers(self):
        return getattr_recursive(self, self.decoder_layers_attr_name)

    def _set_decoder_layers(self, value):
        setattr_recursive(self, self.decoder_layers_attr_name, value)

    def init_flamingo(
        self,
        media_token_id,
        vis_hidden_size,
        cross_attn_every_n_layers,
        use_media_placement_augmentation,
        only_attend_previous: bool,
        device: Optional[Union[str, 'torch.device']] = None,
        dtype: Optional[Union[str, 'torch.dtype']] = None,
    ):
        """
        Initialize Flamingo by adding a new gated cross attn to the decoder. Store the media token id for computing the media locations.
        """

        self.gated_cross_attn_layers = nn.ModuleList(
            [
                GatedCrossAttentionBlock(
                    dim=self.config.hidden_size, dim_visual=vis_hidden_size
                )
                if (layer_idx + 1) % cross_attn_every_n_layers == 0
                else None
                for layer_idx, _ in enumerate(self._get_decoder_layers())
            ]
        )
        self._set_decoder_layers(
            nn.ModuleList(
                [
                    FlamingoLayer(gated_cross_attn_layer, decoder_layer)
                    for gated_cross_attn_layer, decoder_layer in zip(
                        self.gated_cross_attn_layers, self._get_decoder_layers()
                    )
                ]
            )
        )
        self.media_token_id = media_token_id
        self.use_media_placement_augmentation = use_media_placement_augmentation
        self.initialized_flamingo = True
        self.only_attend_previous = only_attend_previous

        dtype, device = auto_dtype_and_device(dtype, device)
        if str(dtype) == 'torch.float16':
            self.gated_cross_attn_layers.half()
        # self.gated_cross_attn_layers.to(device)

    def forward(self, *input, **kwargs):
        """Condition the Flamingo layers on the media locations before forward()"""

        if not self.initialized_flamingo:
            raise ValueError(
                "Flamingo layers are not initialized. Please call `init_flamingo` first."
            )

        input_ids = kwargs["input_ids"] if "input_ids" in kwargs else input[0]
        media_locations = input_ids == self.media_token_id
        # IMPORTANT: Force `attend_previous` to True when we place training data as <image>caption<|endofchunk|>
        # attend_previous = (
        #     (random.random() < 0.5) if self.use_media_placement_augmentation else False
        # )
        attend_previous = self.only_attend_previous

        for layer in self.get_decoder().layers:
            layer.condition_media_locations(media_locations)
            layer.condition_attend_previous(attend_previous)

        return super().forward(
            *input, **kwargs
        )  # Call the other parent's forward method

    def is_conditioned(self) -> bool:
        """Check whether all decoder layers are already conditioned."""
        return all(l.is_conditioned() for l in self._get_decoder_layers())

    def clear_conditioned_layers(self):
        for layer in self._get_decoder_layers():
            layer.condition_vis_x(None)
            layer.condition_media_locations(None)
            layer.condition_attend_previous(None)



from typing import List, Optional, Union

import torch
from loguru import logger

from open_gpt.helper import auto_dtype_and_device


def load_model_and_tokenizer(
    model_name_or_path: str,
    tokenizer_name_or_path: Optional[str] = None,
    device: Optional[torch.device] = None,
    dtype: Optional[Union[str, torch.dtype]] = None,
    device_map: Optional[Union[str, List[int]]] = None,
    no_split_module_classes: Optional[List[str]] = None,
    **kwargs,
):
    """Load a model and tokenizer from HuggingFace."""
    import os

    from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

    dtype, device = auto_dtype_and_device(dtype, device)

    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name_or_path or model_name_or_path, trust_remote_code=True
    )

    if tokenizer.pad_token is None:
        # Issue: GPT models don't have a pad token
        # tokenizer.add_special_tokens({"pad_token": "<PAD>"})
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.pad_token_id = tokenizer.eos_token_id

    # For generation padding tokens should be on the left
    tokenizer.padding_side = "left"

    if device_map:
        import huggingface_hub
        from accelerate import init_empty_weights, load_checkpoint_and_dispatch

        if not os.path.exists(model_name_or_path):
            model_path = huggingface_hub.snapshot_download(model_name_or_path)
        else:
            model_path = model_name_or_path

        config = AutoConfig.from_pretrained(model_name_or_path, trust_remote_code=True)
        with init_empty_weights():
            model = AutoModelForCausalLM.from_config(
                config, torch_dtype=dtype, trust_remote_code=True
            )
            # make sure token embedding weights are still tied if needed
            model.tie_weights()

            model = load_checkpoint_and_dispatch(
                model,
                model_path,
                device_map=device_map,
                no_split_module_classes=no_split_module_classes,
                dtype=dtype,
            )
    else:
        model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path, torch_dtype=dtype, trust_remote_code=True
        )
        model.to(device)

    if hasattr(model, 'generation_config'):
        # set pad_token_id to eos_token_id because GPT does not have a PAD token
        model.generation_config.pad_token_id = model.generation_config.eos_token_id

    return model, tokenizer



from typing import List, Optional, Union

import torch
from tqdm import tqdm

from open_gpt.helper import auto_dtype_and_device
from open_gpt.logging import logger


def load_model_and_tokenizer(
    model_name_or_path: str,
    tokenizer_name_or_path: Optional[str] = None,
    device: Optional[torch.device] = None,
    dtype: Optional[Union[str, torch.dtype]] = None,
    device_map: Optional[Union[str, List[int]]] = None,
    no_split_module_classes: Optional[List[str]] = None,
    **kwargs,
):
    """Load a model and tokenizer from HuggingFace."""
    import os

    from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

    dtype, device = auto_dtype_and_device(dtype, device)

    model_size = model_name_or_path.split('-')[-3]

    facebook_llama_model_name_or_path = f"facebook/llama-{model_size}"
    if os.path.exists(facebook_llama_model_name_or_path):
        llama_model_name_or_path = facebook_llama_model_name_or_path
    else:
        llama_model_name_or_path = f"decapoda-research/llama-{model_size}-hf"

    tokenizer = AutoTokenizer.from_pretrained(
        model_name_or_path or tokenizer_name_or_path, trust_remote_code=True
    )

    if tokenizer.pad_token is None:
        # Issue: GPT models don't have a pad token
        # tokenizer.add_special_tokens({"pad_token": "<PAD>"})
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.pad_token_id = tokenizer.eos_token_id

    # For generation padding tokens should be on the left
    tokenizer.padding_side = "left"

    logger.info(
        f"Loading llama-{model_size} base model from {llama_model_name_or_path}"
    )
    if device_map:
        import huggingface_hub
        from accelerate import init_empty_weights, load_checkpoint_and_dispatch

        if not os.path.exists(llama_model_name_or_path):
            model_path = huggingface_hub.snapshot_download(llama_model_name_or_path)
        else:
            model_path = llama_model_name_or_path

        with init_empty_weights():
            config = AutoConfig.from_pretrained(
                llama_model_name_or_path, trust_remote_code=True
            )
            model = AutoModelForCausalLM.from_config(
                config, torch_dtype=dtype, trust_remote_code=True
            )
            # make sure token embedding weights are still tied if needed
            model.tie_weights()

            model = load_checkpoint_and_dispatch(
                model,
                model_path,
                device_map=device_map,
                no_split_module_classes=no_split_module_classes,
                dtype=dtype,
            )
    else:
        model = AutoModelForCausalLM.from_pretrained(
            llama_model_name_or_path, torch_dtype=dtype, trust_remote_code=True
        )
        model.to(device)

    logger.info(f"Loading model weights delta from {model_name_or_path}")
    delta = AutoModelForCausalLM.from_pretrained(
        model_name_or_path, torch_dtype=torch.float16, low_cpu_mem_usage=True
    )
    # adapted from `fastchat.model.apply_delta`
    for name, param in tqdm(model.state_dict().items(), desc="Applying delta"):
        assert name in delta.state_dict()
        param.data += delta.state_dict()[name].to(param.dtype).to(param.device)

    # clean up delta to save memory
    delta = None
    torch.cuda.empty_cache()

    return model, tokenizer



code_highlight_css = """
#chatbot .hll { background-color: #ffffcc }
#chatbot .c { color: #408080; font-style: italic }
#chatbot .err { border: 1px solid #FF0000 }
#chatbot .k { color: #008000; font-weight: bold }
#chatbot .o { color: #666666 }
#chatbot .ch { color: #408080; font-style: italic }
#chatbot .cm { color: #408080; font-style: italic }
#chatbot .cp { color: #BC7A00 }
#chatbot .cpf { color: #408080; font-style: italic }
#chatbot .c1 { color: #408080; font-style: italic }
#chatbot .cs { color: #408080; font-style: italic }
#chatbot .gd { color: #A00000 }
#chatbot .ge { font-style: italic }
#chatbot .gr { color: #FF0000 }
#chatbot .gh { color: #000080; font-weight: bold }
#chatbot .gi { color: #00A000 }
#chatbot .go { color: #888888 }
#chatbot .gp { color: #000080; font-weight: bold }
#chatbot .gs { font-weight: bold }
#chatbot .gu { color: #800080; font-weight: bold }
#chatbot .gt { color: #0044DD }
#chatbot .kc { color: #008000; font-weight: bold }
#chatbot .kd { color: #008000; font-weight: bold }
#chatbot .kn { color: #008000; font-weight: bold }
#chatbot .kp { color: #008000 }
#chatbot .kr { color: #008000; font-weight: bold }
#chatbot .kt { color: #B00040 }
#chatbot .m { color: #666666 }
#chatbot .s { color: #BA2121 }
#chatbot .na { color: #7D9029 }
#chatbot .nb { color: #008000 }
#chatbot .nc { color: #0000FF; font-weight: bold }
#chatbot .no { color: #880000 }
#chatbot .nd { color: #AA22FF }
#chatbot .ni { color: #999999; font-weight: bold }
#chatbot .ne { color: #D2413A; font-weight: bold }
#chatbot .nf { color: #0000FF }
#chatbot .nl { color: #A0A000 }
#chatbot .nn { color: #0000FF; font-weight: bold }
#chatbot .nt { color: #008000; font-weight: bold }
#chatbot .nv { color: #19177C }
#chatbot .ow { color: #AA22FF; font-weight: bold }
#chatbot .w { color: #bbbbbb }
#chatbot .mb { color: #666666 }
#chatbot .mf { color: #666666 }
#chatbot .mh { color: #666666 }
#chatbot .mi { color: #666666 }
#chatbot .mo { color: #666666 }
#chatbot .sa { color: #BA2121 }
#chatbot .sb { color: #BA2121 }
#chatbot .sc { color: #BA2121 }
#chatbot .dl { color: #BA2121 }
#chatbot .sd { color: #BA2121; font-style: italic }
#chatbot .s2 { color: #BA2121 }
#chatbot .se { color: #BB6622; font-weight: bold }
#chatbot .sh { color: #BA2121 }
#chatbot .si { color: #BB6688; font-weight: bold }
#chatbot .sx { color: #008000 }
#chatbot .sr { color: #BB6688 }
#chatbot .s1 { color: #BA2121 }
#chatbot .ss { color: #19177C }
#chatbot .bp { color: #008000 }
#chatbot .fm { color: #0000FF }
#chatbot .vc { color: #19177C }
#chatbot .vg { color: #19177C }
#chatbot .vi { color: #19177C }
#chatbot .vm { color: #19177C }
#chatbot .il { color: #666666 }
"""
# .highlight  { background: #f8f8f8; }



import dataclasses
from typing import TYPE_CHECKING, List, Optional, Union, overload

import torch
from transformers import GenerationConfig

if TYPE_CHECKING:
    from transformers import AutoModelForCausalLM, AutoTokenizer


class GenerationMixin:
    """Mixin for generation methods."""

    model: 'AutoModelForCausalLM'
    tokenizer: 'AutoTokenizer'

    @overload
    def generate(self, prompt: str, inplace_images: List = [], **kwargs):
        ...

    @overload
    def generate(
        self,
        prompts: Union[str, List[str]],
        max_new_tokens: Optional[int] = None,
        num_beams: int = 1,
        do_sample: bool = False,
        temperature: float = 1.0,
        top_k: int = 1,
        top_p: float = 0.9,
        repetition_penalty: float = 1.0,
        length_penalty: float = 1.0,
        no_repeat_ngram_size: int = 0,
        **kwargs
    ):
        """Generate text from the given prompt.

        :param prompts: The prompt(s) to generate from.
        :param max_new_tokens: The maximum number of tokens to generate, not including the prompt.
        :param num_beams: Number of beams for beam search. 1 means no beam search.
        :param do_sample: Whether to use sampling instead of greedy decoding.
        :param temperature: The temperature to use for sampling. Only relevant if do_sample is True. Higher means more stochastic.
        :param top_k: The number of highest probability vocabulary tokens to keep for top-k-filtering. Only relevant if do_sample is True.
        :param top_p: The cumulative probability of parameter highest probability vocabulary tokens to keep for nucleus sampling. Only relevant if do_sample is True.
        :param repetition_penalty: The parameter for repetition penalty. 1.0 means no penalty.
        :param length_penalty: Exponential penalty to the length that is used with beam-based generation.
                It is applied as an exponent to the sequence length, which in turn is used to divide the score of the sequence.
                Since the score is the log likelihood of the sequence (i.e. negative), length_penalty > 0.0 promotes longer sequences,
                while length_penalty < 0.0 encourages shorter sequences.
        :param no_repeat_ngram_size: If set to int > 0, all ngrams of that size can only occur once.
        """
        ...

    def generate(self, prompts: Union[str, List[str]], **kwargs):
        inputs = self.tokenizer(
            [prompts] if isinstance(prompts, str) else prompts,
            padding=True,
            return_tensors="pt",
        )

        # Move inputs to the correct device
        for k, v in inputs.items():
            if isinstance(v, torch.Tensor):
                inputs[k] = v.to(self._device)

        # overwrite default values with kwargs
        clean_up_tokenization_spaces = kwargs.pop('clean_up_tokenization_spaces', True)
        skip_special_tokens = kwargs.pop("skip_special_tokens", True)

        with torch.inference_mode():
            outputs = self.model.generate(**inputs, **kwargs)

            texts_outs = []
            for _, generated_sequence in enumerate(outputs):
                generated_sequence = generated_sequence.tolist()
                prompt = prompts[_] if isinstance(prompts, list) else prompts
                prompt_len = len(prompt)

                text = self.tokenizer.decode(
                    generated_sequence,
                    clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                    skip_special_tokens=skip_special_tokens,
                )

                text = text[prompt_len:] if text[:prompt_len] == prompt else text
                texts_outs.append(text.lstrip())

            return texts_outs if isinstance(prompts, list) else texts_outs[0]



import logging as _logging
import os as _os
import sys

_logging.captureWarnings(True)


# import importlib.metadata if available, otherwise importlib_metadata
if sys.version_info >= (3, 8):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import jina as _jina

# hide jina user survey
_os.environ['JINA_HIDE_SURVEY'] = '1'


try:
    __jina_version__ = _jina.__version__
except AttributeError as e:
    raise RuntimeError(
        '`jina` dependency is not installed correctly, please reinstall with `pip install -U --force-reinstall jina`'
    )


def get_version() -> str:
    """Return the module version number specified in pyproject.toml.
    :return: The version number.
    """
    return importlib_metadata.version(__package__ + '_torch')


__version__ = get_version()

from .factory import create_model



from typing import Optional

from open_gpt.models.modeling import BaseModel


class PythiaModel(BaseModel):
    """Pythia model.

    It contains two sets of eight models of sizes 70M, 160M, 410M, 1B, 1.4B, 2.8B, 6.9B, and 12B. For each size,
    there are two models: one trained on the Pile, and one trained on the Pile after the dataset has been globally deduplicated.

    See https://github.com/EleutherAI/pythia for more information.

    The quick way to use Vicuna via :meth:`open_gpt.create_model`:

    ```python
    import open_gpt

    model = open_gpt.create_model(
        'EleutherAI/pythia-12b-deduped', precision='fp16', device_map='balanced'
    )
    ```
    """

    no_split_module_classes = ["GPTNeoXLayer"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from typing import Callable, Optional, Union

import torch
from accelerate.hooks import AlignDevicesHook, add_hook_to_module
from einops import rearrange
from open_flamingo.src.helpers import PerceiverResampler
from torch import nn

from open_gpt.logging import logger

from ...helper import auto_dtype_and_device


class FlamingoLMModel(nn.Module):
    def __init__(
        self,
        vision_encoder: 'nn.Module',
        language_model: 'nn.Module',
        device: Optional[Union[str, 'torch.device']] = None,
        dtype: Optional[Union[str, 'torch.dtype']] = None,
        model_config: dict = {},
        **kwargs,
    ):
        """An open source version of DeepMind's Flamingo model!
        Adapted from the original implementation at https://github.com/mlfoundations/open_flamingo

        :param vision_encoder: the vision encoder to extract visual features, e.g. CLIP model
        :param language_model: the language model to extract textual features and generate the output texts, e.g., LLaMa model
        :param model_config: a dictionary of model configuration
        :param device: the device to run the model on
        :param dtype: the data type to run the model on
        :param kwargs: other arguments
        """
        super().__init__()

        self.dtype, self.device = auto_dtype_and_device(dtype, device)

        self.model_config = model_config
        self.vision_encoder = vision_encoder
        self.lang_encoder = language_model

        self.perceiver = PerceiverResampler(dim=self.model_config['image_size'])
        if str(self.dtype) == 'torch.float16':
            self.perceiver.half()
        elif str(self.dtype) == 'torch.int8':
            raise NotImplementedError('int8 is not supported yet')
        self.perceiver.to(self.device)

        self.media_token_id = model_config['media_token_id']
        self.end_chunk_token_id = model_config['end_chunk_token_id']

        logger.debug(
            f"media_token_id: {self.media_token_id}, end_chunk_token_id: {self.end_chunk_token_id}"
        )

        self.lang_encoder.init_flamingo(
            media_token_id=self.media_token_id,
            vis_hidden_size=model_config['image_size'],
            cross_attn_every_n_layers=model_config['cross_attn_every_n_layers'],
            use_media_placement_augmentation=True,
            only_attend_previous=True,
            dtype=dtype,
            device=device,
        )

    def forward(
        self,
        vision_inputs: 'torch.Tensor',
        text_inputs: 'torch.Tensor',
        attention_mask: Optional['torch.Tensor'] = None,
        labels: Optional['torch.Tensor'] = None,
        past_key_values: Optional['torch.Tensor'] = None,
    ):
        """

        :param vision_inputs: vision images input with shape (B, T, F, Channels, Height, Width).
                Images in the same chunk are collated along T, and frames are collated along F
                Currently only F=1 is supported (single-frame videos)
        :param text_inputs: language text input with shape (batch_size, sequence_length)
        :param attention_mask: attention mask, 1 for tokens that are not masked, 0 for masked tokens. Defaults to None.
        :param labels: labels for computing the language modeling loss. Defaults to None.
        :param past_key_values: cached past key and values for fast decoding. Defaults to None.
        """

        self._vision_encode(vision_inputs)

        output = self.lang_encoder(
            input_ids=text_inputs,
            attention_mask=attention_mask,
            labels=labels,
            past_key_values=past_key_values,
            use_cache=False,
        )

        self.lang_encoder.clear_conditioned_layers()

        return output

    def generate(
        self,
        vision_inputs: 'torch.Tensor',
        text_inputs: 'torch.Tensor',
        attention_mask: torch.Tensor = None,
        num_beams: int = 1,
        max_new_tokens: Optional[int] = None,
        temperature: float = 1.0,
        top_k: int = 0,
        top_p: float = 1.0,
        no_repeat_ngram_size: int = 0,
        prefix_allowed_tokens_fn: Optional[Callable] = None,
        length_penalty: float = 1.0,
        num_return_sequences: int = 1,
        do_sample: bool = False,
        early_stopping: bool = True,
    ):
        """
        Generate text conditioned on vision and language inputs.

        :param vision_inputs: vision images input with shape (B, T, F, Channels, Height, Width).
                Images in the same chunk are collated along T, and frames are collated along F
                Currently only F=1 is supported (single-frame videos)
        :param text_inputs: language text input with shape (batch_size, sequence_length)
        :param attention_mask: attention mask, 1 for tokens that are not masked, 0 for masked tokens. Defaults to None.
        :param num_beams: number of beams for beam search. Defaults to 1.
        :param max_new_tokens: maximum number of new tokens to generate. Defaults to None.
        :param temperature: temperature for sampling. Defaults to 1.0.
        :param top_k: top k for sampling. Defaults to 0.
        :param top_p: top p for sampling. Defaults to 1.0.
        :param no_repeat_ngram_size: no repeat ngram size. Defaults to 0.
        :param prefix_allowed_tokens_fn: prefix allowed tokens function. Defaults to None.
        :param length_penalty: length penalty. Defaults to 1.0.
        :param num_return_sequences: number of return sequences. Defaults to 1.
        :param do_sample: whether to sample or not. Defaults to False.
        :param early_stopping: whether to stop early or not. Defaults to False.

        :return: text_inputs with generated tokens appended to it (batch_size, sequence_length)
        """

        # if hasattr(self, "_hf_hook"):
        # add a hook to make sure that the output of lang_encoder is mapped to the same device as the text_inputs
        hook = AlignDevicesHook(
            execution_device=self.device,
            io_same_device=True,
            place_submodules=False,
        )
        add_hook_to_module(self.lang_encoder, hook)

        vision_inputs = vision_inputs.to(dtype=self.dtype, device=self.device)
        # if attention_mask is not None:
        #     attention_mask = attention_mask.to(self.device)
        text_inputs = text_inputs.to(self.device)

        if num_beams > 1:
            vision_inputs = vision_inputs.repeat_interleave(num_beams, dim=0)

        _vision_x = self._vision_encode(vision_inputs=vision_inputs)

        if not self.lang_encoder.initialized_flamingo:
            raise ValueError(
                "Flamingo layers are not initialized. Please call `init_flamingo` first."
            )

        media_locations = text_inputs == self.media_token_id
        attend_previous = False

        for layer in self.lang_encoder.get_decoder().layers:
            layer.condition_media_locations(media_locations)
            layer.condition_attend_previous(attend_previous)

        output = self.lang_encoder.generate(
            text_inputs,
            attention_mask=attention_mask,
            eos_token_id=self.end_chunk_token_id,
            num_beams=num_beams,
            max_new_tokens=max_new_tokens,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            no_repeat_ngram_size=no_repeat_ngram_size,
            length_penalty=length_penalty,
            num_return_sequences=num_return_sequences,
            do_sample=do_sample,
            early_stopping=early_stopping,
        )

        self.lang_encoder.clear_conditioned_layers()
        return output

    def _vision_encode(self, vision_inputs: 'torch.Tensor') -> 'torch.Tensor':
        """
        Compute media tokens from vision input by passing it through vision encoder and conditioning language model.

        :param vision_inputs: vision images input with shape (B, T, F, channels, height, width).
                Images in the same chunk are collated along T, and frames are collated along F
                Currently only F=1 is supported (single-frame videos)
        :return: media tokens with shape (batch_size, frames, tokens, dim)

        rearrange code based on https://github.com/dhansmair/flamingo-mini
        """

        assert (
            vision_inputs.ndim == 6
        ), "vision_inputs should be of shape (B, T, F, Channels, Height, Width)"
        B, T, F = vision_inputs.shape[:3]
        assert F == 1, "Only single frame supported"

        vision_x = rearrange(vision_inputs, "B T F c h w -> (B T F) c h w")

        with torch.no_grad():
            vision_x = self.vision_encoder.visual(vision_x)[1]

        vision_x = rearrange(vision_x, "(B T F) v d -> B T F v d", B=B, T=T, F=F)

        vision_x = self.perceiver(vision_x)  # reshapes to (B, T, n, d)

        for layer in self.lang_encoder._get_decoder_layers():
            layer.condition_vis_x(vision_x)

        return vision_x
